
# ilang - Inference Language 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 


import unittest
from .. import Tracers
from .. import exceptions
from .. import verbose

verbose.set_verbose_low()

class TestSequenceTracer(unittest.TestCase):
    """Sequence of tests for the Tracer sampler"""
    def setUp(self):
        pass
    
    def test_trace(self):
        """.."""
        pass




if __name__ == '__main__':
    unittest.main()



